==> epic2.002.tar.gz: ircii-EPIC2.002/source/ircserv.c <==
/*
 * ircserv.c: A quaint little program to make irc life PING free 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#if 0
static	char	rcsid[] = "@(#)$Id: ircserv.c,v 1.17 1994/07/25 13:36:10 mrg Exp $";
#endif

#include "defs.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif
#include <sys/file.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>

#include "irc_std.h"

/* machines we don't want to use <unistd.h> on 'cause its broken */
#if defined(pyr) || defined(_SEQUENT_)
# undef HAVE_UNISTD_H
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
int	connect_to_unix _((char *));
#endif /* HAVE_SYS_UN_H */

extern	void	set_socket_options _((int));
extern	int	new_select _((fd_set *, fd_set *, struct timeval *));
extern	int	dgets _((char *, int, int, char *));

#ifdef __STDC__
char *new_malloc (int size)
#else
char	*new_malloc(size)
int	size;
#endif
{
	char	*ptr;

	if ((ptr = (char *) malloc(size)) == (char *) 0)
	{
		printf("-1 0\n");
		exit(1);
	}
	return (ptr);
}

/*
 * new_free:  Why do this?  Why not?  Saves me a bit of trouble here and
 * there 
 */
#ifdef __STDC__
void new_free (char **ptr)
#else
void	new_free(ptr)
char	**ptr;
#endif
{
	if (*ptr)
	{
		free(*ptr);
		*ptr = (char *) 0;
	}
}

/*
 * Connect_By_Number Performs a connecting to socket 'service' on host
 * 'host'.  Host can be a hostname or ip-address.  If 'host' is null, the
 * local host is assumed.   The parameter full_hostname will, on return,
 * contain the expanded hostname (if possible).  Note that full_hostname is a
 * pointer to a char *, and is allocated by connect_by_numbers() 
 *
 * Errors: 
 *
 * -1 get service failed 
 *
 * -2 get host failed 
 *
 * -3 socket call failed 
 *
 * -4 connect call failed 
 */
#ifdef __STDC__
int connect_by_number (int service, char *host)
#else
int	connect_by_number(service, host)
int	service;
char	*host;
#endif
{
	int	s;
	char	buf[100];
	struct	sockaddr_in	server;
	struct	hostent	*hp;

	if (host == (char *) 0)
	{
		gethostname(buf, sizeof(buf));
		host = buf;
	}
	if ((server.sin_addr.s_addr = inet_addr(host)) == -1)
	{
		if ((hp = gethostbyname(host)) != NULL)
		{
			bzero((char *) &server, sizeof(server));
			bcopy(hp->h_addr, (char *) &server.sin_addr,
				hp->h_length);
			server.sin_family = hp->h_addrtype;
		}
		else
			return (-2);
	}
	else
		server.sin_family = AF_INET;
	server.sin_port = (unsigned short) htons(service);
	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return (-3);
	set_socket_options(s);
#ifdef PRIV_PORT
	if (geteuid() == 0)
	{
		struct	sockaddr_in	localaddr;
		int	portnum;

		localaddr = server;
		localaddr.sin_addr.s_addr = INADDR_ANY;
		for (portnum = 1023; portnum > PRIV_PORT; portnum--)
		{
			localaddr.sin_port = htons(portnum);
			if (bind(s, (struct sockaddr *) &localaddr,
					sizeof(localaddr)) != -1)
				break;
		}
	}
#endif /*PRIV_PORT*/
	if (connect(s, (struct sockaddr *) &server, sizeof(server)) < 0)
	{
		close(s);
		return (-4);
	}
	return (s);
}

#ifdef HAVE_SYS_UN_H
/*
 * Connect to a UNIX domain socket. Only works for servers.
 * submitted by Avalon for use with server 2.7.2 and beyond.
 */
#ifdef __STDC__
int connect_to_unix (char *path)
#else
int	connect_to_unix(path)
char	*path;
#endif
{
	struct	sockaddr_un	un;
	int	sock;

	sock = socket(AF_UNIX, SOCK_STREAM, 0);

	un.sun_family = AF_UNIX;
	strcpy(un.sun_path, path);
	if (connect(sock, (struct sockaddr *)&un, strlen(path)+2) == -1)
	{
		close(sock);
		return -1;
	}
	return sock;
}
#endif

/*
 * ircserv: This little program connects to the server (given as arg 1) on
 * the given port (given as arg 2).  It then accepts input from stdin and
 * sends it to that server. Likewise, it reads stuff sent from the server and
 * sends it to stdout.  Simple?  Yes, it is.  But wait!  There's more!  It
 * also intercepts server PINGs and automatically responds to them.   This
 * frees up the process that starts ircserv (such as IRCII) to pause without
 * fear of being pooted off the net. 
 */
/* Some problems with the original ircserv:
 *	* Assumption that since you just got a select() call any
 *		write() to the client wont block.
 *	  This obviously is not reliable because it relies on a 
 *		race condition to validate.
 *	* Pinged out after filling up the blocking buffer
 *	  Now it blocks forever simply mallocing happily on the way.
 */
/* This implementation is non-blocking.  What this means
 * is if you suspend your ircII client, ircserv will sit happily
 * reading from the server and answering pings and so forth.  While
 * you are suspended, the client will not make any attempt to write
 * to the ircII client.  When you resume the client, ircserv will
 * dump everything on the client ASAP.  The memory used to buffer 
 * the stuff coming from the server is allocated dynamically, which
 * means you can be suspended forever without pinging out at the
 * expense of some ram. 
 *
 * (Written/Modified by jnelson@rush.cc.edu for the EPIC project)
 */
int main (int argc, char *argv[])
{
	int	des;			/* server fd */

	/* stuff from the for(;;) loop */
	fd_set	rd,			/* the ircII client + server */
		wd;			/* the ircII client */
	int	wrote,			/* how much we wrote */
		c;			/* length of blocking buffer */
	char	buffer[BUFSIZ + 1],	/* buffer for incoming stuff */
		*block_buffer = NULL,	/* stuff thats blocked */
		*quit_message = "QUIT :ircserv caught EOF!\r\n",
		*pong = "PONG . .";

	/* must have 3 args to make any sense */
	if (argc < 3)
	{
		fprintf(stderr, "Usage: %s server port\n", argv[0]);
		exit(1);
	}

#ifdef HAVE_SYS_UN_H
        if (*argv[1] == '/')
                des = connect_to_unix(argv[1]);
        else
#endif /* HAVE_SYS_UN_H */
                des = connect_by_number(atoi(argv[2]), argv[1]);

	/* if we couldnt connect, punt */
        if (des < 0)
                exit(des);

	(void) MY_SIGNAL(SIGTERM, SIG_IGN, 0);
        (void) MY_SIGNAL(SIGSEGV, SIG_IGN, 0);
        (void) MY_SIGNAL(SIGPIPE, SIG_IGN, 0);
#ifdef SIGWINCH
        (void) MY_SIGNAL(SIGWINCH, SIG_IGN, 0);
#endif
#ifdef SIGBUS
        (void) MY_SIGNAL(SIGBUS, SIG_IGN, 0);
#endif

	/* mark stdout (ircII client) as nonblocking */
        fflush(stdout);
        if (fcntl(1, F_SETFL, FNDELAY))
                exit(1);

	/* Each iteration through this loop is one read cycle */
	/* Loop runs forever until explicitly stopped */
	for (;;)
	{
		fflush(stderr);

		/* set up our fd's every run through */
		FD_ZERO(&rd);
		FD_SET(0, &rd);
		FD_SET(des, &rd);
		FD_ZERO(&wd);

		/* If we are waiting for the client to unblock,
		 * then we set it for writing, otherwise we leave
		 * it as uninteresting.
		 */
		if (block_buffer)
			FD_SET(1, &wd);

		switch (new_select(&rd, &wd, NULL))
		{
		case -1:
		case  0:	break;

		default:
		{
			/* we can write to the client */
			if (FD_ISSET(1, &wd))
			{
				c = strlen(block_buffer);

				/* The old ircserv made the assumption that
				   since we just got out of a select() call
				   that if write() returns -1 that we got
				   an EOF.  This is bogus because nothing
				   guarantees us that just becuase we could
				   write a few ms ago we can still write now
				   (ie, race condition).  So we *must* always
				   check for EAGAIN before we give up
				 */
				if ((wrote = write(1, block_buffer, c)) == -1)
				{
					/* something happened to ircII */
					if (errno != EAGAIN)
					{
						/* We were sent an EOF 
						   if the client didnt bother
						   to clean up, we do it here */
						write(des, quit_message, strlen(quit_message));
						exit(0);
					}
				}

				else if (wrote < c)
					/* since we're not blocking, we just 
					   write whatever we write and save the 
					   rest for later... no biggie */
					strcpy(block_buffer, &(block_buffer[wrote]));
			}

			/* stuff incoming from the client */
			if (FD_ISSET(0, &rd))
			{
				/* dump it right out to the server. no prob. */
				if (c = dgets(buffer, BUFSIZ, 0, (char *) 0))
					write(des, buffer, c);
				else
				{
					write(des, quit_message, strlen(quit_message));
					exit(0);
				}
			}

			/* incoming stuff from the server */
			if (FD_ISSET(des, &rd))
			{
				/* get the line */
				if ((c = dgets(buffer, BUFSIZ, des, (char *) 0)) != -1)
				{
					/* wheeeee. */
					buffer[c] = 0;

					/* if its a ping, handle it */
					if (strncmp(buffer, "PING ", 5) == 0)
						write(des, pong, strlen(pong));

					/* its a regular line */
					else
					{
						/* tack it to the end of the stuff
						   we've already blocked on */
						int foo = strlen(block_buffer) + strlen(buffer);
						char *tmp = (char *)new_malloc(foo);
						strcpy(tmp, block_buffer);
						strcat(tmp, buffer);
						new_free(&block_buffer);
						block_buffer = tmp;

						/* try to write it to the client */
						if ((wrote = write(1, block_buffer, c)) == -1)
						{
							/* if it didnt work but
							   we didnt block, its
							   and EOF from the
							   client. wrap up */
							if (errno != EAGAIN)
							{
								write(des, quit_message, strlen(quit_message));
								exit(0);
							}
							/* if we DO get EAGAIN
							   here, then we WANT to keep
							   the buffer as-is since no
							   bytes were written. */
						}

						/* we wrote some, but not all of the buffer. save the rest 
						 * I am making a calculated guess -- on the knowledge c cannot
						 * be negative here, i strcpy instead of m_strcpy to save some
						 * ram and a lot of time in malloc -- so even if we write some
						 * but not all of the buffer, the actual size of the block
						 * doesnt decrease, we just dont use it all.
						 */
						else if (wrote < c)
							strcpy(block_buffer, &block_buffer[wrote]);

						/* we wrote the whole buffer */
						else
						{
							new_free(&block_buffer);
							block_buffer = NULL;
						}
					}
				}
				/* EOF from server */
				else
					exit(0);
			} 
		} 
		}
	}
}

==> epic2.003.tar.gz: ircii-EPIC2.003/source/ircserv.c <==
/*
 * ircserv.c: A quaint little program to make irc life PING free 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#if 0
static	char	rcsid[] = "@(#)$Id: ircserv.c,v 1.17 1994/07/25 13:36:10 mrg Exp $";
#endif

#include "defs.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif
#include <sys/file.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>

#include "irc_std.h"

/* machines we don't want to use <unistd.h> on 'cause its broken */
#if defined(pyr) || defined(_SEQUENT_)
# undef HAVE_UNISTD_H
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
int	connect_to_unix _((char *));
#endif /* HAVE_SYS_UN_H */

extern	void	set_socket_options _((int));
extern	int	new_select _((fd_set *, fd_set *, struct timeval *));
extern	int	dgets _((char *, int, int, char *));

#ifdef __STDC__
char *new_malloc (int size)
#else
char	*new_malloc(size)
int	size;
#endif
{
	char	*ptr;

	if ((ptr = (char *) malloc(size)) == (char *) 0)
	{
		printf("-1 0\n");
		exit(1);
	}
	return (ptr);
}

/*
 * new_free:  Why do this?  Why not?  Saves me a bit of trouble here and
 * there 
 */
#ifdef __STDC__
void new_free (char **ptr)
#else
void	new_free(ptr)
char	**ptr;
#endif
{
	if (*ptr)
	{
		free(*ptr);
		*ptr = (char *) 0;
	}
}

/*
 * Connect_By_Number Performs a connecting to socket 'service' on host
 * 'host'.  Host can be a hostname or ip-address.  If 'host' is null, the
 * local host is assumed.   The parameter full_hostname will, on return,
 * contain the expanded hostname (if possible).  Note that full_hostname is a
 * pointer to a char *, and is allocated by connect_by_numbers() 
 *
 * Errors: 
 *
 * -1 get service failed 
 *
 * -2 get host failed 
 *
 * -3 socket call failed 
 *
 * -4 connect call failed 
 */
#ifdef __STDC__
int connect_by_number (int service, char *host)
#else
int	connect_by_number(service, host)
int	service;
char	*host;
#endif
{
	int	s;
	char	buf[100];
	struct	sockaddr_in	server;
	struct	hostent	*hp;

	if (host == (char *) 0)
	{
		gethostname(buf, sizeof(buf));
		host = buf;
	}
	if ((server.sin_addr.s_addr = inet_addr(host)) == -1)
	{
		if ((hp = gethostbyname(host)) != NULL)
		{
			bzero((char *) &server, sizeof(server));
			bcopy(hp->h_addr, (char *) &server.sin_addr,
				hp->h_length);
			server.sin_family = hp->h_addrtype;
		}
		else
			return (-2);
	}
	else
		server.sin_family = AF_INET;
	server.sin_port = (unsigned short) htons(service);
	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return (-3);
	set_socket_options(s);
#ifdef PRIV_PORT
	if (geteuid() == 0)
	{
		struct	sockaddr_in	localaddr;
		int	portnum;

		localaddr = server;
		localaddr.sin_addr.s_addr = INADDR_ANY;
		for (portnum = 1023; portnum > PRIV_PORT; portnum--)
		{
			localaddr.sin_port = htons(portnum);
			if (bind(s, (struct sockaddr *) &localaddr,
					sizeof(localaddr)) != -1)
				break;
		}
	}
#endif /*PRIV_PORT*/
	if (connect(s, (struct sockaddr *) &server, sizeof(server)) < 0)
	{
		close(s);
		return (-4);
	}
	return (s);
}

#ifdef HAVE_SYS_UN_H
/*
 * Connect to a UNIX domain socket. Only works for servers.
 * submitted by Avalon for use with server 2.7.2 and beyond.
 */
#ifdef __STDC__
int connect_to_unix (char *path)
#else
int	connect_to_unix(path)
char	*path;
#endif
{
	struct	sockaddr_un	un;
	int	sock;

	sock = socket(AF_UNIX, SOCK_STREAM, 0);

	un.sun_family = AF_UNIX;
	strcpy(un.sun_path, path);
	if (connect(sock, (struct sockaddr *)&un, strlen(path)+2) == -1)
	{
		close(sock);
		return -1;
	}
	return sock;
}
#endif

/*
 * ircserv: This little program connects to the server (given as arg 1) on
 * the given port (given as arg 2).  It then accepts input from stdin and
 * sends it to that server. Likewise, it reads stuff sent from the server and
 * sends it to stdout.  Simple?  Yes, it is.  But wait!  There's more!  It
 * also intercepts server PINGs and automatically responds to them.   This
 * frees up the process that starts ircserv (such as IRCII) to pause without
 * fear of being pooted off the net. 
 */
/* Some problems with the original ircserv:
 *	* Assumption that since you just got a select() call any
 *		write() to the client wont block.
 *	  This obviously is not reliable because it relies on a 
 *		race condition to validate.
 *	* Pinged out after filling up the blocking buffer
 *	  Now it blocks forever simply mallocing happily on the way.
 */
/* This implementation is non-blocking.  What this means
 * is if you suspend your ircII client, ircserv will sit happily
 * reading from the server and answering pings and so forth.  While
 * you are suspended, the client will not make any attempt to write
 * to the ircII client.  When you resume the client, ircserv will
 * dump everything on the client ASAP.  The memory used to buffer 
 * the stuff coming from the server is allocated dynamically, which
 * means you can be suspended forever without pinging out at the
 * expense of some ram. 
 *
 * (Written/Modified by jnelson@rush.cc.edu for the EPIC project)
 */
int main (int argc, char *argv[])
{
	int	des;			/* server fd */

	/* stuff from the for(;;) loop */
	fd_set	rd,			/* the ircII client + server */
		wd;			/* the ircII client */
	int	wrote,			/* how much we wrote */
		c;			/* length of blocking buffer */
	char	buffer[BUFSIZ + 1],	/* buffer for incoming stuff */
		*block_buffer = NULL,	/* stuff thats blocked */
		*quit_message = "QUIT :ircserv caught EOF!\r\n",
		*pong = "PONG . .";

	/* must have 3 args to make any sense */
	if (argc < 3)
	{
		fprintf(stderr, "Usage: %s server port\n", argv[0]);
		exit(1);
	}

#ifdef HAVE_SYS_UN_H
        if (*argv[1] == '/')
                des = connect_to_unix(argv[1]);
        else
#endif /* HAVE_SYS_UN_H */
                des = connect_by_number(atoi(argv[2]), argv[1]);

	/* if we couldnt connect, punt */
        if (des < 0)
                exit(des);

#if 0 	/* why? makes it hard to debug... */
	(void) MY_SIGNAL(SIGTERM, SIG_IGN, 0);
        (void) MY_SIGNAL(SIGSEGV, SIG_IGN, 0);
        (void) MY_SIGNAL(SIGPIPE, SIG_IGN, 0);
#ifdef SIGWINCH
        (void) MY_SIGNAL(SIGWINCH, SIG_IGN, 0);
#endif
#ifdef SIGBUS
        (void) MY_SIGNAL(SIGBUS, SIG_IGN, 0);
#endif
#endif

	/* mark stdout (ircII client) as nonblocking */
        fflush(stdout);
        if (fcntl(1, F_SETFL, FNDELAY))
                exit(1);

	/* Each iteration through this loop is one read cycle */
	/* Loop runs forever until explicitly stopped */
	for (;;)
	{
		fflush(stderr);

		/* set up our fd's every run through */
		FD_ZERO(&rd);
		FD_SET(0, &rd);
		FD_SET(des, &rd);
		FD_ZERO(&wd);

		/* If we are waiting for the client to unblock,
		 * then we set it for writing, otherwise we leave
		 * it as uninteresting.
		 */
		if (block_buffer)
			FD_SET(1, &wd);

		switch (new_select(&rd, &wd, NULL))
		{
		case -1:
		case  0:	break;

		default:
		{
			/* we can write to the client */
			if (FD_ISSET(1, &wd))
			{
				c = strlen(block_buffer);

				/* The old ircserv made the assumption that
				   since we just got out of a select() call
				   that if write() returns -1 that we got
				   an EOF.  This is bogus because nothing
				   guarantees us that just becuase we could
				   write a few ms ago we can still write now
				   (ie, race condition).  So we *must* always
				   check for EAGAIN before we give up
				 */
				if ((wrote = write(1, block_buffer, c)) == -1)
				{
					/* something happened to ircII */
					if (errno != EAGAIN)
					{
						/* We were sent an EOF 
						   if the client didnt bother
						   to clean up, we do it here */
						write(des, quit_message, strlen(quit_message));
						exit(0);
					}
				}

				else if (wrote < c)
					/* since we're not blocking, we just 
					   write whatever we write and save the 
					   rest for later... no biggie */
					strcpy(block_buffer, &(block_buffer[wrote]));
				else
				{
					new_free(&block_buffer);
					block_buffer = NULL;
				}
			}

			/* stuff incoming from the client */
			if (FD_ISSET(0, &rd))
			{
				/* dump it right out to the server. no prob. */
				if (c = dgets(buffer, BUFSIZ, 0, (char *) 0))
					write(des, buffer, c);
				else
				{
					write(des, quit_message, strlen(quit_message));
					exit(0);
				}
			}

			/* incoming stuff from the server */
			if (FD_ISSET(des, &rd))
			{
				/* get the line */
				if ((c = dgets(buffer, BUFSIZ, des, (char *) 0)) != -1)
				{
					/* wheeeee. */
					buffer[c] = 0;

					/* if its a ping, handle it */
					if (strncmp(buffer, "PING ", 5) == 0)
						write(des, pong, strlen(pong));

					/* its a regular line */
					else
					{
						/* tack it to the end of the stuff
						   we've already blocked on */
						char *tmp;
						int foo = 0;
						if (block_buffer) 
							foo = strlen(block_buffer);
						foo += strlen(buffer) + 1;
						tmp = (char *)new_malloc(foo);
						*tmp = 0;
						if (block_buffer)
							strcpy(tmp, block_buffer);
						strcat(tmp, buffer);
						new_free(&block_buffer);
						block_buffer = tmp;

						/* try to write it to the client */
						if ((wrote = write(1, block_buffer, c)) == -1)
						{
							/* if it didnt work but
							   we didnt block, its
							   and EOF from the
							   client. wrap up */
							if (errno != EAGAIN)
							{
								write(des, quit_message, strlen(quit_message));
								exit(0);
							}
							/* if we DO get EAGAIN
							   here, then we WANT to keep
							   the buffer as-is since no
							   bytes were written. */
						}

						/* we wrote some, but not all of the buffer. save the rest 
						 * I am making a calculated guess -- on the knowledge c cannot
						 * be negative here, i strcpy instead of m_strcpy to save some
						 * ram and a lot of time in malloc -- so even if we write some
						 * but not all of the buffer, the actual size of the block
						 * doesnt decrease, we just dont use it all.
						 */
						else if (wrote < c)
							strcpy(block_buffer, &block_buffer[wrote]);

						/* we wrote the whole buffer */
						else
						{
							new_free(&block_buffer);
							block_buffer = NULL;
						}
					}
				}
				/* EOF from server */
				else
					exit(0);
			} 
		} 
		}
	}
}
